@@ -28,61 +28,25 @@
 ///        In-depth description of this algorithm:
 ///        https://github.com/kimwalisch/primecount/blob/master/doc/Hard-Special-Leaves.md
 ///
-/// Copyright (C) 2022 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2025 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
 ///
 
-#include <Sieve.hpp>
-#include <SieveTables.hpp>
+#include "Sieve.hpp"
+#include "Sieve_arrays.hpp"
+#include "Sieve_count_start_stop.hpp"
+#include "Sieve_pre_sieve.hpp"
+
 #include <imath.hpp>
+#include <int128_t.hpp>
 #include <macros.hpp>
 #include <min.hpp>
-#include <pod_vector.hpp>
-#include <popcnt.hpp>
 
 #include <stdint.h>
 #include <algorithm>
-
-using std::fill_n;
-using std::sqrt;
-using primecount::pod_array;
-
-namespace {
-
-struct WheelInit
-{
-  uint8_t factor;
-  uint8_t index;
-};
-
-/// Categorize sieving primes according to their modulo 30
-/// congruence class { 1, 7, 11, 13, 17, 19, 23, 29 }.
-///
-const pod_array<uint8_t, 30> wheel_offsets =
-{
-  0, 8 * 0, 0, 0, 0, 0,
-  0, 8 * 1, 0, 0, 0, 8 * 2,
-  0, 8 * 3, 0, 0, 0, 8 * 4,
-  0, 8 * 5, 0, 0, 0, 8 * 6,
-  0, 0,     0, 0, 0, 8 * 7
-};
-
-/// Used to calculate the first multiple > start of a
-/// sieving prime that is coprime to 2, 3, 5.
-///
-const pod_array<WheelInit, 30> wheel_init
-{{
-  {1,  0}, {0,  0}, {5,  1}, {4,  1}, {3,  1},
-  {2,  1}, {1,  1}, {0,  1}, {3,  2}, {2,  2},
-  {1,  2}, {0,  2}, {1,  3}, {0,  3}, {3,  4},
-  {2,  4}, {1,  4}, {0,  4}, {1,  5}, {0,  5},
-  {3,  6}, {2,  6}, {1,  6}, {0,  6}, {5,  7},
-  {4,  7}, {3,  7}, {2,  7}, {1,  7}, {0,  7}
-}};
-
-} // namespace
+#include <cmath>
 
 namespace primecount {
 
@@ -94,14 +58,13 @@ Sieve::Sieve(uint64_t low,
   ASSERT(segment_size % 240 == 0);
 
   start_ = low;
-  segment_size = get_segment_size(segment_size);
+  segment_size = align_segment_size(segment_size);
 
   // sieve_size = segment_size / 30 as each byte corresponds
   // to 30 numbers i.e. the 8 bits correspond to the
   // offsets = {1, 7, 11, 13, 17, 19, 23, 29}.
   sieve_.resize(segment_size / 30);
   wheel_.reserve(wheel_size);
-  wheel_.resize(4);
   allocate_counter(low);
 }
 
@@ -118,26 +81,27 @@ Sieve::Sieve(uint64_t low,
 ///
 void Sieve::allocate_counter(uint64_t low)
 {
-  double average_leaf_dist = sqrt(low);
-  double counter_dist = sqrt(average_leaf_dist);
+  double average_leaf_dist = std::sqrt(low);
+  double counter_dist = std::sqrt(average_leaf_dist);
 
   // Here we balance counting with the counter array and
-  // counting from the sieve array using the POPCNT
-  // instruction. Since the POPCNT instructions allows to
-  // count a distance of 240 using a single instruction we
-  // slightly increase the counter distance and slightly
-  // decrease the size of the counter array.
-  counter_.dist = (uint64_t) (counter_dist * sqrt(240));
-
-  // Each byte represents an interval of size 30
-  uint64_t bytes = counter_.dist / 30;
+  // counting from the sieve array using the 64-bit POPCNT
+  // instruction. Since the 64-bit POPCNT instructions
+  // allows to count a distance of 240 using a single
+  // instruction we slightly increase the counter distance
+  // and slightly decrease the size of the counter array.
+  uint64_t bytes_count_instruction = bytes_per_count_instruction();
+  ASSERT(bytes_count_instruction >= sizeof(uint64_t));
+  uint64_t dist_per_instruction = bytes_count_instruction * 30;
+  counter_.dist = uint64_t(counter_dist * std::sqrt(dist_per_instruction));
 
   // Increasing the minimum counter distance decreases the
-  // branch mispredictions but on the other hand increases
-  // the number of executed instructions. On newer CPUs
-  // reducing the branch mispredictions is more important
-  // than reducing the number of executed instructions.
-  bytes = max((uint64_t) sizeof(uint64_t) * 16, bytes);
+  // branch mispredictions (good) but on the other hand
+  // increases the number of executed instructions (bad).
+  // In my benchmarks setting the minimum amount of bytes to
+  // bytes_count_instruction * 8 (or 16) performed best.
+  uint64_t bytes = counter_.dist / 30;
+  bytes = max(bytes, bytes_count_instruction * 8);
   bytes = next_power_of_2(bytes);
 
   // Make sure the counter (32-bit) doesn't overflow.
@@ -145,7 +109,7 @@ void Sieve::allocate_counter(uint64_t low)
   // only counts the number of unsieved elements (1 bits) in
   // an interval of size: sieve_limit^(1/4) * sqrt(240).
   // Hence the max(counter value) = 2^18.
-  ASSERT(bytes * 8 <= std::numeric_limits<uint32_t>::max());
+  ASSERT(bytes * 8 <= pstd::numeric_limits<uint32_t>::max());
   uint64_t counter_size = ceil_div(sieve_.size(), bytes);
   counter_.counter.resize(counter_size);
   counter_.dist = bytes * 30;
@@ -164,7 +128,7 @@ uint64_t Sieve::segment_size() const
 /// process 64-bit words (8 bytes) and each
 /// byte contains 30 numbers.
 ///
-uint64_t Sieve::get_segment_size(uint64_t size)
+uint64_t Sieve::align_segment_size(uint64_t size)
 {
   size = max(size, 240);
 
@@ -174,15 +138,15 @@ uint64_t Sieve::get_segment_size(uint64_t size)
   return size;
 }
 
-void Sieve::reset_sieve(uint64_t low, uint64_t high)
+/// Use shorter sieve size for last segment
+void Sieve::resize_sieve(uint64_t low, uint64_t high)
 {
-  fill_n(sieve_.data(), sieve_.size(), 0xff);
   uint64_t size = high - low;
 
   if (size < segment_size())
   {
     uint64_t last = size - 1;
-    size = get_segment_size(size);
+    size = align_segment_size(size);
     sieve_.resize(size / 30);
     auto sieve64 = (uint64_t*) sieve_.data();
     sieve64[last / 240] &= unset_larger[last % 240];
@@ -220,82 +184,33 @@ void Sieve::init_counter(uint64_t low, uint64_t high)
   }
 }
 
-/// Count 1 bits inside [0, stop]
-uint64_t Sieve::count(uint64_t stop)
-{
-  ASSERT(stop >= prev_stop_);
-  uint64_t start = prev_stop_ + 1;
-  prev_stop_ = stop;
-
-  // Quickly count the number of unsieved elements (in
-  // the sieve array) up to a value that is close to
-  // the stop number i.e. (stop - start) < counter_.dist.
-  // We do this using the counter array, each element
-  // of the counter array contains the number of
-  // unsieved elements in the interval:
-  // [i * counter_.dist, (i + 1) * counter_.dist[.
-  while (counter_.stop <= stop)
-  {
-    start = counter_.stop;
-    counter_.stop += counter_.dist;
-    counter_.sum += counter_[counter_.i++];
-    count_ = counter_.sum;
-  }
-
-  // Here the remaining distance is relatively small i.e.
-  // (stop - start) < counter_.dist, hence we simply
-  // count the remaining number of unsieved elements by
-  // linearly iterating over the sieve array.
-  count_ += count(start, stop);
-  return count_;
-}
-
-/// Count 1 bits inside [start, stop]
-uint64_t Sieve::count(uint64_t start, uint64_t stop) const
-{
-  if (start > stop)
-    return 0;
-
-  ASSERT(stop - start < segment_size());
-
-  uint64_t start_idx = start / 240;
-  uint64_t stop_idx = stop / 240;
-  uint64_t m1 = unset_smaller[start % 240];
-  uint64_t m2 = unset_larger[stop % 240];
-  auto sieve64 = (uint64_t*) sieve_.data();
-
-  if (start_idx == stop_idx)
-    return popcnt64(sieve64[start_idx] & (m1 & m2));
-  else
-  {
-    uint64_t cnt = popcnt64(sieve64[start_idx] & m1);
-    for (uint64_t i = start_idx + 1; i < stop_idx; i++)
-      cnt += popcnt64(sieve64[i]);
-    cnt += popcnt64(sieve64[stop_idx] & m2);
-    return cnt;
-  }
-}
-
 /// Add a sieving prime to the sieve.
 /// Calculates the first multiple > start of prime that
 /// is not divisible by 2, 3, 5 and its wheel index.
 ///
-void Sieve::add(uint64_t prime)
+void Sieve::add(uint64_t prime, uint64_t i)
 {
-  ASSERT(start_ % 30 == 0);
+  if_unlikely(i > wheel_.size())
+    wheel_.resize(i);
 
-  // first multiple > start_
+  // Find first multiple > start_
+  ASSERT(start_ % 30 == 0);
   uint64_t quotient = start_ / prime + 1;
   uint64_t multiple = prime * quotient;
 
-  // find next multiple of prime that
+  // Find next multiple of prime that
   // is not divisible by 2, 3, 5
   uint64_t factor = wheel_init[quotient % 30].factor;
   multiple += prime * factor;
+
+  ASSERT(multiple % 2 != 0);
+  ASSERT(multiple % 3 != 0);
+  ASSERT(multiple % 5 != 0);
+
   multiple = (multiple - start_) / 30;
   uint32_t multiple32 = (uint32_t) multiple;
 
-  // calculate wheel index of multiple
+  // Calculate wheel index of multiple
   uint32_t index = wheel_init[quotient % 30].index;
   index += wheel_offsets[prime % 30];
   wheel_.emplace_back(multiple32, index);
@@ -307,7 +222,7 @@ void Sieve::add(uint64_t prime)
 void Sieve::cross_off(uint64_t prime, uint64_t i)
 {
   if (i >= wheel_.size())
-    add(prime);
+    add(prime, i);
 
   prime /= 30;
   Wheel& wheel = wheel_[i];
@@ -323,6 +238,8 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
       return; \
     }
 
+  ASSERT(wheel.index <= 63);
+
   switch (wheel.index)
   {
     for (;;)
@@ -563,7 +480,7 @@ void Sieve::cross_off(uint64_t prime, uint64_t i)
 void Sieve::cross_off_count(uint64_t prime, uint64_t i)
 {
   if (i >= wheel_.size())
-    add(prime);
+    add(prime, i);
 
   reset_counter();
   Wheel& wheel = wheel_[i];
@@ -588,13 +505,14 @@ void Sieve::cross_off_count(uint64_t prime, uint64_t i)
   #define COUNT_UNSET_BIT(bit_index) \
     { \
       std::size_t sieve_byte = sieve[m]; \
-      std::size_t bit = sieve_byte & (1 << bit_index); \
-      std::size_t is_bit = bit >> bit_index; \
-      sieve[m] = (uint8_t) (sieve_byte - bit); \
+      std::size_t is_bit = (sieve_byte >> bit_index) & 1; \
+      sieve[m] &= ~(1 << bit_index); \
       counter[m >> counter_log2_dist] -= (uint32_t) is_bit; \
-      total_count -= is_bit; \
+      total_count -= (uint64_t) is_bit; \
     }
 
+  ASSERT(wheel.index <= 63);
+
   switch (wheel.index)
   {
     for (;;)
